/*

    multimap -- C/python module providing C++/STL multimaps

    Copyright (C) 2003  Eric S. Tiedemann

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact: Eric S. Tiedemann <est@hyperreal.org>

*/

/*
  TODO
  more multimap ops: eq, lt, keys, values, has_item
  tests of: find
  use as-map interface?
  gc support
  reverse iterators
  iterator equality comparison
  priority queue use example
  use getitem, setitem and delitem?
  reinterpret vs static cast?
  catch and propagate C++ bad_alloc exceptions
  way to erase a particular key/value item
  evalable repr
  double multimap (specialized to use doubles as keys)
*/

#include <Python.h>

#include <map>

using namespace std;

// Let's have C++ handle reference counting and comparison of Python
// objects in our C++ containers via this PyObjectXX class.
class PyObjectXX {
public:
  PyObject *o;

  PyObjectXX(PyObject *);
  PyObjectXX(const PyObjectXX &);
  PyObjectXX &operator=(const PyObjectXX &);
  ~PyObjectXX();
  bool operator <(const PyObjectXX &that) const;
  bool operator ==(const PyObjectXX &that) const;
};

inline PyObjectXX::PyObjectXX(PyObject *o) : o(o) {
  Py_INCREF(o);
}

inline PyObjectXX::PyObjectXX(const PyObjectXX &that) : o(that.o) {
  Py_INCREF(o);
}

inline PyObjectXX &PyObjectXX::operator=(const PyObjectXX &that) {
  Py_DECREF(o);
  o = that.o;
  Py_INCREF(o);
  return *this;
}

inline PyObjectXX::~PyObjectXX() {
  Py_DECREF(o);
}

inline bool PyObjectXX::operator <(const PyObjectXX &that) const {
  return PyObject_Compare(o, that.o) == -1;
}

inline bool PyObjectXX::operator ==(const PyObjectXX &that) const {
  return PyObject_Compare(o, that.o) == 0;
}

// A couple of typedefs to make the rest of the file prettier.
typedef multimap<PyObjectXX, PyObjectXX> PyObjMultiMap;
typedef PyObjMultiMap::const_iterator PyObjMultiMapItr;


// Our Python MultiMap object.
typedef struct
{
  PyObject_HEAD
  PyObjMultiMap mm;
  int read_only;
} MultiMap;

#define MultiMap_Check(o) ((o)->ob_type == &MultiMap_Type)


// We're going to need a Python Object for iterators too.
typedef struct
{
  PyObject_HEAD
  PyObject *mm;
  PyObjMultiMapItr itr;
} MultiMapIterator;

#define MultiMapIterator_Check(o) ((o)->ob_type == &MultiMapIterator_Type)

static PyObject *
MultiMapIterator_next(PyObject *pyself)
{
  MultiMapIterator *self = reinterpret_cast<MultiMapIterator *>(pyself);
  MultiMap *mm = reinterpret_cast<MultiMap *>(self->mm);

  if (self->itr == mm->mm.end()) {
    return NULL;
  }

  PyObject *key = self->itr->first.o;
  PyObject *value = self->itr->second.o;
  PyObject *item = PyTuple_New(2);
  if (!item) {
    return NULL;
  }
  Py_INCREF(key);
  Py_INCREF(value);
  PyTuple_SET_ITEM(item, 0, key);
  PyTuple_SET_ITEM(item, 1, value);

  self->itr++;

  return item;
}

static void
MultiMapIterator_del(PyObject *self)
{
  MultiMapIterator *itr = reinterpret_cast<MultiMapIterator *>(self);
  itr->itr.~PyObjMultiMapItr();
  reinterpret_cast<MultiMap *>(itr->mm)->read_only--;
  Py_DECREF(itr->mm);
  PyObject_Del(self);
}

static PyTypeObject MultiMapIterator_Type =
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "MultiMapIterator",
  sizeof(MultiMapIterator),
  0,

  (destructor) MultiMapIterator_del,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  PyObject_GenericGetAttr,		/* tp_getattro */
  0,
  0,
  Py_TPFLAGS_DEFAULT,			/* tp_flags */

  0,
  0,
  0,

  0,
  0,
  PyObject_SelfIter,			/* tp_iter */
  MultiMapIterator_next,	        /* tp_iternext */
};

static PyObject *
MultiMapIterator_new(MultiMap *map, const PyObjMultiMapItr &itr)
{
  MultiMapIterator *myitr = PyObject_NEW(MultiMapIterator, 
					 &MultiMapIterator_Type);
  if (!myitr)
    return NULL;
  new (&myitr->itr) PyObjMultiMapItr(itr);
  Py_INCREF(map);
  myitr->mm = reinterpret_cast<PyObject *>(map);
  reinterpret_cast<MultiMap *>(myitr->mm)->read_only++;

  return reinterpret_cast<PyObject *>(myitr);
}

// end of iterator code


static PyObject *
MultiMap_insert(PyObject *self, PyObject *args)
{
  PyObject *key, *val;

  if (!PyArg_ParseTuple(args, "OO:MultiMap.insert", &key, &val))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);

  if (mm->read_only) {
    PyErr_SetString(PyExc_RuntimeError, 
		    "insert into MultiMap while iterator exists");
    return NULL;
  }

  mm->mm.insert(pair<PyObjectXX, PyObjectXX>(key, val));

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
MultiMap_clear(PyObject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":MultiMap.clear"))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);

  if (mm->read_only) {
    PyErr_SetString(PyExc_RuntimeError, 
		    "clear() of MultiMap while iterator exists");
    return NULL;
  }

  mm->mm.clear();

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
MultiMap_erase(PyObject *self, PyObject *args)
{
  PyObject *key;

  if (!PyArg_ParseTuple(args, "O:MultiMap.erase", &key))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);

  if (mm->read_only) {
    PyErr_SetString(PyExc_RuntimeError, 
		    "erase() on MultiMap while iterator exists");
    return NULL;
  }

  mm->mm.erase(PyObjectXX(key));

  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
MultiMap_popmin(PyObject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":MultiMap.popmin"))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);

  if (mm->read_only) {
    PyErr_SetString(PyExc_RuntimeError, 
		    "popmin() on MultiMap while iterator exists");
    return NULL;
  }

  if (mm->mm.empty()) {
    PyErr_SetString(PyExc_IndexError, "popmin() on empty MultiMap");
    return NULL;
  }

  PyObjMultiMap::iterator p = mm->mm.begin();

  PyObject *key = p->first.o;
  PyObject *value = p->second.o;
  PyObject *item = PyTuple_New(2);
  if (!item) {
    return NULL;
  }
  Py_INCREF(key);
  Py_INCREF(value);
  PyTuple_SET_ITEM(item, 0, key);
  PyTuple_SET_ITEM(item, 1, value);

  mm->mm.erase(p);

  return item;
}

static PyObject *
MultiMap_items(PyObject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":MultiMap.items"))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);

  // No need to check for map mutations since we hold the GIL.
  PyObject *lst = PyList_New(mm->mm.size());
  if (!lst)
    return NULL;
  
  int lstidx = 0;
  for (PyObjMultiMapItr p = mm->mm.begin();
       p != mm->mm.end();
       ++p) {
    PyObject *key = p->first.o;
    PyObject *value = p->second.o;
    PyObject *item = PyTuple_New(2);
    if (!item) {
      Py_DECREF(lst);
      return NULL;
    }
    Py_INCREF(key);
    Py_INCREF(value);
    PyTuple_SET_ITEM(item, 0, key);
    PyTuple_SET_ITEM(item, 0, key);
    PyTuple_SET_ITEM(item, 1, value);
    PyList_SET_ITEM(lst, lstidx++, item);
  }

  return lst;
}

static PyObject *
MultiMap_iteritems(PyObject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ":MultiMap.iteritems"))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);

  return MultiMapIterator_new(mm, mm->mm.begin());
}

// Our implicit iterator method is the same as the iteritems() method.
static PyObject *
MultiMap_iter(PyObject *self)
{
  MultiMap *mm = reinterpret_cast<MultiMap *>(self);

  return MultiMapIterator_new(mm, mm->mm.begin());
}

static PyObject *
MultiMap_find(PyObject *self, PyObject *args)
{
  PyObject *key;

  if (!PyArg_ParseTuple(args, "O:MultiMap.find", &key))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);
  
  return MultiMapIterator_new(mm, mm->mm.find(PyObjectXX(key)));
}

static PyObject *
MultiMap_has_key(PyObject *self, PyObject *args)
{
  PyObject *key;

  if (!PyArg_ParseTuple(args, "O:MultiMap.has_key", &key))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);
  
  PyObjMultiMapItr p = mm->mm.find(PyObjectXX(key));

  long ok = (p != mm->mm.end() && p->first == PyObjectXX(key));

  return PyBool_FromLong(ok);
}

static PyObject *
MultiMap_count(PyObject *self, PyObject *args)
{
  PyObject *key;

  if (!PyArg_ParseTuple(args, "O:MultiMap.count", &key))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);
  
  long count = mm->mm.count(PyObjectXX(key));

  return PyInt_FromLong(count);
}

static PyObject *
MultiMap_lower_bound(PyObject *self, PyObject *args)
{
  PyObject *key;

  if (!PyArg_ParseTuple(args, "O:MultiMap.lower_bound", &key))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);
  
  return MultiMapIterator_new(mm, mm->mm.lower_bound(PyObjectXX(key)));
}

static PyObject *
MultiMap_upper_bound(PyObject *self, PyObject *args)
{
  PyObject *key;

  if (!PyArg_ParseTuple(args, "O:MultiMap.upper_bound", &key))
    return NULL;

  MultiMap *mm = reinterpret_cast<MultiMap *>(self);

  return MultiMapIterator_new(mm, mm->mm.upper_bound(PyObjectXX(key)));
}

static struct PyMethodDef MultiMap_methods[] = {
  {"insert", MultiMap_insert, METH_VARARGS,
   "insert(key, value) -> insert an item with key KEY and value VALUE"},
  {"clear", MultiMap_clear, METH_VARARGS, "clear() -> erase all entries"},
  {"items", MultiMap_items, METH_VARARGS,
   "items() -> <list of key/value pairs>"},
  {"iteritems", MultiMap_iteritems, METH_VARARGS,
   "iteritems() -> <iterator on key/value pairs>"},
  {"find", MultiMap_find, METH_VARARGS,
   "find(key) -> <iterator at first item with key KEY>"},
  {"has_key", MultiMap_has_key, METH_VARARGS,
   "has_key(key) -> <boolean indicating whether KEY has an entry>"},
  {"count", MultiMap_count, METH_VARARGS,
   "count(key) -> <count of entries with key KEY>"},
  {"erase", MultiMap_erase, METH_VARARGS,
   "erase(key) -> erase all items with key KEY"},
  {"popmin", MultiMap_popmin, METH_VARARGS,
   "popmin() -> <key and value for first item with lowest key>\n"
   "  Also erases the item returned.\n"
   "  Useful for priority-queue style usage.\n"
   "  Raises IndexError if multimap is empty."},
  {"lower_bound", MultiMap_lower_bound, METH_VARARGS, 
   "lower_bound(key) -> <iterator at first item with key not less than KEY>"},
  {"upper_bound", MultiMap_upper_bound, METH_VARARGS,
   "upper_bound(key) -> <iterator at first item with key greater than KEY>"},
  {NULL, NULL, 0, NULL}
};

static int
MultiMap_length(MultiMap *p)
{
  return p->mm.size();
}

static PyObject *
MultiMap_getattr(PyObject *self, char *name)
{
  if (strcmp(name, "__methods__") == 0)
    return Py_BuildValue("[sssssssssss]",
                         "insert", "clear", "items",
                         "iteritems", "find", "has_key", "count", 
			 "erase", "popmin",
                         "lower_bound", "upper_bound"
                         );
  else if (strcmp(name, "__members__") == 0)
    return Py_BuildValue("[s]", "__doc__");
  else
    return Py_FindMethod(MultiMap_methods, self, name);
}

static void
MultiMap_del(PyObject *self)
{
  MultiMap *mm = reinterpret_cast<MultiMap *>(self);
  mm->mm.~PyObjMultiMap();
  PyObject_Del(self);
}

static PySequenceMethods MultiMap_as_sequence = {
  (inquiry)MultiMap_length, /*sq_length*/
};

PyDoc_STRVAR(multimap_doc,
"MultiMap([<list of pairs>]) -> <multimap object>\n"
"\n"
"MultiMaps are sorted maps in which a key can have multiple values.\n"
"\n"
"Methods:\n"
"  insert(key, value) -> insert an item with key KEY and value VALUE\n"
"  clear() -> erase all entries\n"
"  items() -> <list of key/value pairs>\n"
"  iteritems() -> <iterator on key/value pairs>\n"
"  find(key) -> <iterator at first item with key KEY>\n"
"  has_key(key) -> <boolean indicating whether KEY has an entry>\n"
"  count(key) -> <count of entries with key KEY>\n"
"  erase(key) -> erase all items with key KEY\n"
"  popmin() -> <key and value for first item with lowest key>\n"
"    Also erases the item returned.\n"
"    Useful for priority-queue style usage.\n"
"    Raises IndexError if multimap is empty.\n"
"  lower_bound(key) -> iterator at first item with key not less than KEY\n"
"  upper_bound(key) -> iterator at first item with key greater than KEY\n"
"\n"
"The default iterator iterates through the key/value pairs.\n"
"\n"
"A MultiMap can not be modified while an iterator for it exists.\n"
);

static PyTypeObject MultiMap_Type =
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,
  "MultiMap",
  sizeof(MultiMap),
  0,

  (destructor) MultiMap_del,
  0,
  MultiMap_getattr,
  0,
  0,
  0,

  0,
  &MultiMap_as_sequence,      /*tp_as_sequence*/
  0,

  0,
  0,
  0,

  0,
  0,
  0,
  Py_TPFLAGS_DEFAULT,
  multimap_doc,
  0,
  0,
  0,
  0,
  MultiMap_iter,
};

static bool check_init(PyObject *init) {
  if (!PyList_Check(init))
    return false;

  int n = PyList_GET_SIZE(init);
  for (int i = 0; i < n; i++) {
    PyObject *item = PyList_GET_ITEM(init, i);
    if (!PyTuple_Check(item))
      return false;
    if (PyTuple_GET_SIZE(item) != 2)
      return false;
  }

  return true;
}

static PyObject *
MultiMap_new(PyObject *self, PyObject *args)
{
  PyObject *init = 0;

  if (!PyArg_ParseTuple(args, "|O:MultiMap", &init))
    return NULL;

  if (init && !check_init(init)) {
    PyErr_SetString(PyExc_ValueError, 
		    "MultiMap initializer is not a list of pairs");
    return NULL;
  }

  MultiMap *mymultimap = PyObject_NEW(MultiMap, &MultiMap_Type);
  if (!mymultimap)
    return NULL;
  new (&mymultimap->mm) PyObjMultiMap;
  mymultimap->read_only = 0;

  if (init) {
    int n = PyList_GET_SIZE(init);
    for (int i = 0; i < n; i++) {
      PyObject *item = PyList_GET_ITEM(init, i);
      mymultimap->mm.insert(pair<PyObjectXX, PyObjectXX>
			    (PyTuple_GET_ITEM(item, 0),
			     PyTuple_GET_ITEM(item, 1)));
    }
  }

  return reinterpret_cast<PyObject *>(mymultimap);
}

static struct PyMethodDef
multimapmod_methods[] = {
	{"MultiMap",    MultiMap_new,   METH_VARARGS, multimap_doc},
	{NULL, NULL}
};

extern "C" {
DL_EXPORT(void)
initmultimap(void)
{
	Py_InitModule("multimap", multimapmod_methods);
}
}
